

/**
 * Simple complex number class.
 * A complex number z = (x,y) has both a real component x and
 * an imaginary component y.  Often z is written as x +iy where
 * i is the solution to v^2 = -1.
 *
 *
 * @author Brent Heeringa
 * @version 1.0
 */
public class Complex {
	
	private double _real;
	private double _imaginary;

	/**
	 * Default constructor creates the complex number z = (0,0)
	 */
	public Complex() {
		this(0,0);
	}

	/**
	 * Creates the complex number z = (<tt>real</tt>, <tt>imaginary</tt>)
	 */
	public Complex(double real, double imaginary) {
		_real = real;
		_imaginary = imaginary;
	}
	
	/**
	 * Adding two complex numbers (a,b) (c,d) is defined as
	 * (a+c,b+d)
	 * @returns <tt>this</tt> + <tt>c</tt>
	 */
	public Complex plus(Complex c) {
		return(new Complex(_real + c.getReal(), _imaginary + c.getImaginary()));
	}
	
	/**
	 * Returns the real component of the complex number
	 * @returns the real value
	 */
	public double getReal() {
		return(_real);
	}
	
	/**
	 * Returns the imaginary component of the complex number
	 * @returns the imaginary value
	 */
	public double getImaginary() {
		return(_imaginary);
	}
	
	/**
	 * Subtracting two complex numbers (a,b) (c,d) is defined as
	 * (a - c, b -d)
	 * @returns <tt>this</tt> - <tt>c</tt>
	 */
	public Complex minus(Complex c) {
		return(new Complex(_real - c.getReal(), 
											 _imaginary - c.getImaginary()));
	}
	
	/**
	 * Multiplying two complex numbers (a,b) (c,d) is defined as
	 * (a*c - b*d, a*d + b*c)
	 * @returns <tt>this</tt> * <tt>c</tt>
	 */
	public Complex times(Complex c) {
		return(new Complex(_real * c.getReal() - _imaginary * c.getImaginary(),
											 _real * c.getImaginary() + _imaginary * c.getReal()));
	}

	/**
	 * Dividing two complex numbers z = (a,b) w = (c,d) is defined as
	 * z * w-bar / |w|^2
	 *
	 * The |w|^2 is equivalent to w * w.conjugate() where 
	 * the conjugate of w = (c,d) is defined as (c,-d)
	 * Note that multiplying by the conjugate produces a non-negative real
	 * number so that we can easily divide the real and imaginary
	 * components of the imaginary numbers.
	 * @returns <tt>this</tt> / <tt>c</tt>
	 * @throws ArithmeticException
	 */
	public Complex divide(Complex c) throws ArithmeticException {
		Complex numerator = this.times(c.conjugate());
		Complex denominator = c.times(c.conjugate());
		if (denominator.getReal() == 0) {
			throw(new ArithmeticException());
		}
		return(new Complex(numerator.getReal() / denominator.getReal(),
											 numerator.getImaginary() / denominator.getReal()));
	}
	
	/**
	 * Return the exponant of the complex number.
	 * Note that when power = 0 we return (1,0)
	 * and when power < 0 we return 1 / (a,b)^power
	 * @returns <tt>this^power</tt>
	 */
	public Complex power(int power) {
		Complex c = this;
		if (power == 0) {
			return(new Complex(1.0, 0));
		} else if (power > 0) {
			for (int i = 1; i < power; ++i) {
				c = c.times(this);
			}
			return(c);
		} else {
			c = c.power(-power);
			return((new Complex(1,0)).divide(c));
		}		
	}

	/**
	 * For any complex number z = (a,b) the conjugate 
	 * (usually denoted z-bar) is defined as (a,-b)
	 * @returns The conjugate of <tt>this</tt>
	 */	
	public Complex conjugate() {
		return(new Complex(_real, -_imaginary));
	}
		
	/**
	 * Two complex numbers are equal if and only if both their
	 * real and imaginary components are equal
	 * @returns true if <tt>this</tt> is equivalent to <tt>c</tt>
	 */
	public boolean equals(Complex c) {
		return(_real == c.getReal() &&
					 _imaginary == c.getImaginary()); 	
	}

		
	/**
	 * Two complex numbers are equal if both their
	 * real and imaginary components are equal
	 * @returns true if <tt>this</tt> is equivalent to <tt>o</tt>
	 */
	public boolean equals(Object o) {
		if (!(o instanceof Complex)) {
			return(false);
		} else {
			return(this.equals((Complex) o));
		}	
	}
	
	/**
	 * Returns <pre>(r,i)</pre> where <tt>r</tt> is the real
	 * component and <tt>i</tt> is the imaginary component
	 * @returns String representation of <tt>this</tt>
	 */
	public String toString() {
		return("(" + _real + "," + _imaginary + ")");
	}
	
	/**
	 * Clone procedure
	 * @returns A deep copy of <tt>this</tt>
	 */
	public Object clone() {
	 	return(new Complex(_real, _imaginary));
	}
	
	/**
	 * Tests the Complex API
	 */
	public static void main(String[] args) {
		
		Complex c1 = new Complex();
		Complex c2 = new Complex(0,1);
		Complex c3 = new Complex(1,0);
		Complex c4 = new Complex(4.0,-2.5);

		System.out.println("Hypothesis: (0.0,0.0) Result: " + c1);
		System.out.println("Hypothesis: (0.0,1.0) Result: " + c1.plus(c2));
		System.out.println("Hypothesis: (1.0,-1.0) Result: " + c3.minus(c2));
		System.out.println("Hypothesis: (0.0,1.0) Result: " + c3.times(c2));
		System.out.println("Hypothesis: (9.75,-20.0) Result: " + c4.power(2));
		System.out.println("Hypothesis: (0.0,-1.0) Result: " + c3.divide(c2));
		System.out.println("Hypothesis: false Result: " + c3.equals(c2));
		System.out.println("Hypothesis: true Result: " + c3.equals(c3.clone()));		
		System.out.println("Hypothesis: false Result: " + c3.equals(new Integer(0)));
	}	
}
